"""
Bash printing with font.
"""

import sys


class Status:
    """
    Status indicators.
    """

    UNKNOWN = -1
    SUCCESS = 0
    ERROR = 1
    WARNING = 2


class BackgroundColours:
    """
    Colour codes to highlight text in bash output.
    """

    HEADER = '\033[95m'
    BLUE = '\033[94m'
    CYAN = '\033[96m'
    SUCCESS = '\033[92m'
    WARNING = '\033[93m'
    ERROR = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BACK_TO_DEFAULT = '\033[0m'


def message_with_colour(message, colour, bold=False, underline=False):
    """
    Annotated a message with colour code and highlight codes for bash printing.

    Colour, bold and underline code are added to the front of the message. The back
    to default code is added to the end of the message to not affect printing that
    follows the message.

    :param message: str
        The message.
    :param colour: BackgroundColours constant
        The colour code.
    :param bold: bool.
        If True adds bold code.
    :param underline: bool.
        If True adds underline code.
    :return: str
        The message with added codes.
    """
    font = colour
    if bold:
        font += BackgroundColours.BOLD
    if underline:
        font += BackgroundColours.UNDERLINE
    coloured_message = font + str(message) + BackgroundColours.BACK_TO_DEFAULT
    return coloured_message


def print_to_error(message):
    """
    Send message to stderr.

    :param message: str
        The message
    """
    print(message, file=sys.stderr)


def print_warning(message):
    """
    Prints a message in yellow warning colour.

    :param message: str
        The message.
    """
    print(message_with_colour(message, BackgroundColours.WARNING))


def print_error(message):
    """
    Prints a message in red error colour.

    :param message: str
        The message.
    """
    print_to_error(message_with_colour(message, BackgroundColours.ERROR))


def print_success(message):
    """
    Prints a message in green success colour.

    :param message: str
        The message
    """
    print(message_with_colour(message, BackgroundColours.SUCCESS))


def print_status(message, status, status_text=None):
    """
    Prints a status message in the form: <message>\t[<status>].

    The first part of the status message is any message passed as a parameter.
    Following this with a tab distance and in boxed brackets is a status message
    text in a colour based on the specified status:
        - Status.SUCCESS:
            - text: "SUCCESS"
            - colour: BackgroundColours.SUCCESS (green)
        - Status.ERROR:
            - text: "FAILURE"
            - colour: BackgroundColours.ERROR (red)
        - Status.WARNING:
            - text: "WARNING"
            - colour: BackgroundColours.WARNING (yellow)
        - Unknown/else
            - text: "UNKNOWN"
            - colour: BackgroundColours.BLUE (blue)
    The default message can be overridden by specifying the additional parameter.

    :param message: str
        The message.
    :param status: Status constant
        The status.
    :param status_text: str
        If the value is not None this text overrides the default status message in the
        boxed brackets.
    """
    if status == Status.SUCCESS:
        if status_text is None:
            status_text = "SUCCESS"
        status_colour = BackgroundColours.SUCCESS
    elif status == Status.ERROR:
        if status_text is None:
            status_text = "FAILURE"
        status_colour = BackgroundColours.ERROR
    elif status == Status.WARNING:
        if status_text is None:
            status_text = "WARNING"
        status_colour = BackgroundColours.WARNING
    else:
        if status_text is None:
            status_text = "UNKNOWN"
        status_colour = BackgroundColours.BLUE
    status_indicator = message_with_colour(status_text, status_colour)
    print(message, "\t", f"[{status_indicator}]")



"""
Search for common language errors in a file that can be determined via regex.
"""

import logging
import re


logger = logging.getLogger(__name__)


class ErrorList:
    """
    Class that represent a list of text errors.
    """

    def __init__(self, list_of_errors=None):
        if list_of_errors is None:
            self._errors = []
        else:
            self._errors = list_of_errors

    def __len__(self):
        return len(self._errors)

    def __getitem__(self, i):
        return self._errors[i]

    def __delitem__(self, i):
        del self._errors[i]

    def __setitem__(self, i, val):
        self._errors[i] = val

    def __str__(self):
        return str(self._errors)

    def __repr__(self):
        str_repr = [(e.error_message, e.regex, e.exclude_beginning_words, e.split_characters, e.identifier)
                    for e in self._errors]
        return str_repr

    @property
    def errors(self):
        """
        :return: list
        A list of the errors.
        """
        return self._errors

    @property
    def identifiers(self):
        """
        :return: [str]
        List of all non None error identifiers.
        """
        identifier_list = [e.identifier for e in self._errors if e.identifier is not None]
        return identifier_list

    def get_error_for_identifier(self, identifier):
        """
        Returns the text errors with the specified identifier
        :param identifier: str
        The identifier to search for.
        :return: TextErrorType
        The error if an error with the identifier exists. Otherwise None.
        """
        found_error = next((e for e in self._errors if e.identifier == identifier), None)
        return found_error

    @classmethod
    def to_yaml(cls, dumper, data):
        """
        Transform an object of ErrorList into a MappingNode from a list containing all the TextErrorTypes
        transformed into dicts.

        :param dumper:
        :param data: ErrorList
        The ErrorList object.
        :return: MappingNode
        The mapping node from the dict of the object.
        """
        error_list = []
        for one_error in data.errors:
            one_error_as_dict = {"message": one_error.error_message,
                                 "regex": one_error.regex.pattern}
            if len(one_error.exclude_beginning_words) > 0:
                one_error_as_dict["exclude_words"] = one_error.exclude_beginning_words
            if one_error.split_characters != " ":
                one_error_as_dict["split_character"] = one_error.split_characters
            if one_error.identifier is not None:
                one_error_as_dict["identifier"] = one_error.identifier
            error_list.append(one_error_as_dict)
        return dumper.represent_mapping("ErrorList", {"errors": error_list})

    @classmethod
    def from_yaml(cls, loader, _, data):
        """
        A MappingNode representing a ErrorList to an ErrorList.

        :param loader:
        :param _:
        :param data: MappingNode
        The MappingNode.
        :return:
        An instance of ErrorList.
        """
        error_list = []
        try:
            list_of_errors_as_nodes = data.value[0][1].value
        except TypeError:
            logger.error("Structural error in language_errors. No errors loaded.")
            list_of_errors_as_nodes = []
        for one_error_node in list_of_errors_as_nodes:
            error_as_dict = loader.construct_mapping(one_error_node)
            if "message" in error_as_dict and "regex" in error_as_dict:
                error_constructor_args = {"error_message": error_as_dict["message"], "regex": error_as_dict["regex"]}
                keys_to_include_if_possible = [("exclude_words", "exclude_beginning_words"),
                                               ("split_character", "split_characters"),
                                               ("identifier", "identifier")]
                for yaml_dict_key, param_name in keys_to_include_if_possible:
                    if yaml_dict_key in error_as_dict:
                        error_constructor_args[param_name] = error_as_dict[yaml_dict_key]
                error_list.append(TextErrorType(**error_constructor_args))
            else:
                warn_message = "A text error could not be loaded."
                if "message" in error_as_dict:
                    warn_message += f" Message '{error_as_dict['message']}'."
                else:
                    warn_message += " Message missing."
                if "regex" in error_as_dict:
                    warn_message += f" Regex '{error_as_dict['regex']}'."
                else:
                    warn_message += " Regex missing."
                logger.error(warn_message)
        return cls(error_list)

    def insert(self, i, val):
        self._errors.insert(i, val)

    def append(self, val):
        self.insert(len(self._errors), val)


class TextErrorType:
    """
    A type of textual error that can be found with a regex.
    """

    def __init__(self, error_message, regex, exclude_beginning_words=None, split_characters=" ", identifier=None):
        """
        Constructor.

        :param error_message: str
        The message for this error.
        :param regex: raw str
        The regex for this error.
        :param exclude_beginning_words: [str]
        List of match's first words for which to ignore match.
        :param split_characters: str
        The character to spit a match to determine the first word.
        :param identifier: str
        A string to identify the error with.
        """
        self.regex = re.compile(regex)
        self.error_message = error_message
        if exclude_beginning_words is None:
            self.exclude_beginning_words = []
        else:
            self.exclude_beginning_words = exclude_beginning_words
        self.split_characters = split_characters
        self.identifier = identifier

    def check_for_error(self, text):
        """
        Search for an occurrence of the error (regex) in the text.

        :param text: str
            The text to search in.
        :return:
        TextErrorOccurrence or None
            None if no error found or the (first) occurrence of the error.
        """
        check_result = self.regex.search(text)
        error = None
        if check_result:
            occurrence_start_idx = check_result.start()
            occurrence_end_idx = check_result.end()
            error = TextErrorOccurrence(text, self, occurrence_start_idx, occurrence_end_idx)
            text_split_at_space = re.split(self.split_characters, error.get_matched_text())
            first_word = text_split_at_space.pop(0)
            if first_word in self.exclude_beginning_words:
                error = None
        return error


class TextErrorOccurrence:
    """
    Information on an error occurrence in a text.
    """

    def __init__(self, text, error_type, start_idx_in_text, end_idx_in_text):
        self.error_type = error_type
        self.text = text
        self.start_idx_in_text = start_idx_in_text
        self.end_idx_in_text = end_idx_in_text

    def get_matched_text(self):
        """
        The substring that was found to contain the error.
        :return:
        str
            The substring.
        """
        last_space_before_occurrence = self.last_space_before_occurrence()
        if last_space_before_occurrence == -1:
            last_space_before_occurrence = 0
        else:
            last_space_before_occurrence += 1
        first_space_after_occurrence = self.first_space_after_occurrence()
        if first_space_after_occurrence == -1:
            first_space_after_occurrence = len(self.text) - 1
        matched_text = self.text[last_space_before_occurrence:first_space_after_occurrence]
        return matched_text

    def last_space_before_occurrence(self):
        """
        The index of the last space before the substring where the error occurred.
        :return:
        int
            The index.
        """
        return self.text.rfind(" ", 0, self.start_idx_in_text)

    def first_space_after_occurrence(self):
        """
        The index of the first space after the substring where the error occurred.
        :return:
        int
            The index.
        """
        return self.text.find(" ", self.end_idx_in_text)

    def __repr__(self):
        """
        A textual representation of the error (error message).

        Structure: ~TextErrorType.error_message ' - Character location: from ' <start index of substring>
                   ' to ' <end index of substring>'.\n\t -->' <'...'|'(beginning of line)'> <matched text>
                   <'...'|'(end of line)'>

        :return:
        str
            The error message
        """
        last_space_before_occurrence = self.last_space_before_occurrence()
        before_string = "..."
        if last_space_before_occurrence == -1:
            before_string = "(beginning of line)"
        first_space_after_occurrence = self.first_space_after_occurrence
        after_string = "..."
        if first_space_after_occurrence == -1:
            after_string = "(end of line)"
        message = (self.error_type.error_message
                   + " - Character location: from " + str(self.start_idx_in_text)
                   + " to " + str(self.end_idx_in_text) + "."
                   + "\n\t --> "
                   + before_string
                   + self.get_matched_text()
                   + after_string)
        return message


def filter_errors_list(errors, deactivate=None):
    """
    Excludes the errors identified by the identifier in the deactivate list from the list of all errors.

    :param errors: [TextErrorType]
        List of all errors.
    :param deactivate: [str]
        List of identifiers or errors to not check for.
    """
    errors_to_check_for = errors
    if deactivate is not None:
        errors_to_check_for = [e for e in errors if e.identifier is None or e.identifier not in deactivate]
    return errors_to_check_for


def check_for_error_and_print(file_to_check, errors):
    """
    Checks a file for all known textual errors except those excluded.

    :param file_to_check:
        The path of the file to check for errors.
    :param errors: {int: [TextErrorType]}
        Dict of lists of all errors indexed by the line number.
    """
    i = 0
    all_errors = {}
    with open(file_to_check, "r") as file_to_check:
        for line in file_to_check:
            i += 1
            errors_in_line = []
            for one_error_type in errors:
                error_from_error_type = one_error_type.check_for_error(line)
                if error_from_error_type is not None:
                    errors_in_line.append(error_from_error_type)
            if len(errors_in_line) > 0:
                all_errors[i] = errors_in_line
    return all_errors



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

"""
Latex suite entry script. Includes all argument definitions and the branching class to respective functions.
"""

import logging
import os
import sys
import time
from pathlib import Path

import yaml

import argparse as ap
import argcomplete

import latex_suite.bibliography as bibliography
from latex_suite.git import GitInteraction
from latex_suite import search_language_errors
from latex_suite import latex
from latex_suite import bash_print as bash
import latex_suite.util as util
from latex_suite.util import Configuration, filename_stem

CONF_FILE = "latex_suite.yaml"
LOG_FILE = "latex_suite.log"
NO_TEX_FOUND_WARNING = "No_compatible_tex_files_found"


# logging.basicConfig(filename='example.log', encoding='utf-8', level=logging.INFO)
logging.basicConfig(filename=LOG_FILE, filemode="w",
                    format=f"[%(levelname)s][%(name)s]"
                           + f"[{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}] %(message)s")


class ArgumentParserExtensions:
    """
    Functions for use with the argument parser to parse arguments.
    """

    @staticmethod
    def is_valid_file(file_path, parser):
        """
        An ArgumentParser argument type to check if a file exists.

        Checks if the file with the given path exists. If the file does not exists or if the argument is
        the special value of :py:const:`~NO_TEX_FOUND_WARNING` calls a parser error.

        :param file_path: str
            A path to a file or :py:const:`~NO_TEX_FOUND_WARNING`.
        :param parser:
            The calling sub-parser for an error call if necessary.
        :return:
        str
            The value of file_path if it is an existing file.
        """
        if file_path == NO_TEX_FOUND_WARNING:
            parser.error("No compatible tex file found.")
        elif not os.path.isfile(file_path):
            parser.error(f"The file {file_path} does not exist!")
        else:
            return file_path

    @staticmethod
    def is_valid_directory(path, parser):
        """
        An ArgumentParser argument type to check if a directory exists.

        Checks if the directory with the given path exists. If the directory does not exists calls a parser error.

        :param path: str
            A directory path.
        :param parser:
            The calling sub-parser for an error call if necessary.
        :return:
        str
            The value of path if it is an existing file.
        """
        path = os.path.expanduser(path)
        if not os.path.isdir(path):
            parser.error(f"The folder {path} does not exist!")
        else:
            return path


def make_pdf_with_substitute_defaults(file_to_translate, num_compile_attempts,
                                      verbose=False, do_bib=False, do_single_run=False):
    """
    Translate a pdf using the default pdf engine and bib engine.

    :param file_to_translate:
        the tex file to translate.
    :param num_compile_attempts: int
        The maximal number of attempts to try to pass past a latex translation error.
    :param verbose: bool
        If True print processing.
    :param do_bib: bool
        If True processes the citations with the citation engine.
    :param do_single_run: bool
        If True and do_bib is False one translation of the tex file.
        If True and do_bib is True: tex translation, citation processing, tex translation.
    :return: LatexBashCompileResult
        The compile result of the last compile with the outcome, output and number of errors and warnings.
        If the bibliography was processed the bib processing information is included.
    """
    engine = Configuration.get_config().engine
    bib_engine = Configuration.get_config().bib_engine
    result = latex.make_pdf(file_to_translate, engine, bib_engine,
                            max_end_attempts=num_compile_attempts,
                            verbose=verbose, do_bib=do_bib, do_single_run=do_single_run)
    return result


def load_settings():
    """
    Load the configuration or applies defaults.

    If a yaml configuration is found the file is loaded and used to initiate the configuration. If
    no file is found all default values are used.
    :return: Configuration
        The configuration.
    """
    loaded_config = None
    if os.path.isfile(CONF_FILE):
        with open(CONF_FILE, "r") as conf_file:
            yaml.SafeLoader.add_multi_constructor("ErrorList", search_language_errors.ErrorList.from_yaml)
            loaded_config = yaml.safe_load(conf_file)
    return Configuration(loaded_config)


def compile_one_tex(tex_file, do_bib, is_verbose, num_compile_attempts):
    """
    Compile on tex file.

    If the tex file could not be found or not compiled the program will exit.

    :param tex_file:
        The path of the tex file.
    :param do_bib: bool
        If True processes the citations with the citation engine.
    :param is_verbose: bool
        If True print processing.
    :param num_compile_attempts: int
        The maximal number of attempts to try to pass past a latex translation error.
    """
    compile_result = make_pdf_with_substitute_defaults(tex_file,
                                                       num_compile_attempts=num_compile_attempts,
                                                       verbose=is_verbose, do_bib=do_bib)
    if compile_result.outcome == latex.Outcome.ABORTED:
        bash.print_error("Failed to compile tex file '" + str(tex_file) + "'.")
        sys.exit(1)
    elif compile_result.outcome == latex.Outcome.FILE_NOT_FOUND:
        bash.print_error("Tex file to compile '" + str(tex_file) + "' not found.")
        sys.exit(2)


def task_write_conf(parsed_args):
    """
    Creates a config file with the default configuration in the current directory.

    :param parsed_args:
        Arguments from the command line.
    """

    force_creation = parsed_args["f"]
    if not force_creation and os.path.isfile(CONF_FILE):
        bash.print_error(f"Config file '{CONF_FILE}' already exists. If you want to override it please specify the"
                         + f" force action flag '-f'.")
    else:
        Configuration.write_default(CONF_FILE)


def task_clean_up(parsed_args):
    """
    Removes all files with file extensions in the clean up list.

    :param parsed_args: dict
        The parameter from the arg parser.
    """
    file_extensions = Configuration.get_config().clean_extensions
    depth = parsed_args["d"]
    list_only = parsed_args["l"]
    force_remove = parsed_args["f"]
    num_files, num_files_removed = util.remove_files(file_extensions, directory=".", depth=depth,
                                                     only_list=list_only, force_remove=force_remove)
    if list_only:
        print(f"Found {num_files} file(s).")
    else:
        print(f"Removed {num_files_removed} of {num_files} file(s).")


def task_git(parsed_args):
    """
    Initiates a git related action as specified by the args.

    :param parsed_args:dict
        The parameter from the arg parser.
    """
    action = parsed_args["action"]
    if action == "ignore":
        if not os.path.isfile(".gitignore"):
            gitignore_file = Path(".gitignore")
            gitignore_file.touch()
        extensions_with_wildcard = ["*" + extension for extension in Configuration.get_config().clean_extensions]
        util.add_files_and_extensions_to_file(extensions_with_wildcard, ".gitignore")
    elif action == "credentialUsername":
        username = parsed_args["username"]
        set_username_parameter = GitInteraction.get_git_config_credential_set_username_parameter_string(username)
        errors = GitInteraction.execute_bash_git_cmd(GitInteraction.CONFIG, set_username_parameter)
        if errors is not None:
            sys.exit(errors)


def task_make_bib_file(parsed_args):
    """
    Creates a bib file under the specifications of the cmd args.

    :param parsed_args: dict
        The parameter from the arg parser.
    """
    aux_file_path = parsed_args["auxFile"]
    bib_folder_path = parsed_args["bibsFolder"]
    out_file_path = parsed_args["output"]
    bib_fields_to_remove = Configuration.get_config().bibliography_fields_to_remove
    errors = bibliography.create_bib_file(aux_file_path, bib_folder_path, out_file_path, bib_fields_to_remove)
    for one_error in errors:
        bash.print_error(str(one_error))


def task_lang_command(parsed_args):
    """
    Searches for textual language error in a file as specified by the cmd args.

    :param parsed_args:dict
        The parameter from the arg parser.
    """
    file_to_check = parsed_args["file"]
    exclude = None
    if any(identifier in parsed_args for identifier in Configuration.get_config().language_errors.identifiers):
        exclude = []
        for identifier in Configuration.get_config().language_errors.identifiers:
            if identifier in parsed_args and parsed_args[identifier]:
                exclude.append(identifier)
    all_errors = Configuration.get_config().language_errors
    errors = search_language_errors.filter_errors_list(all_errors, exclude)
    found_errors = search_language_errors.check_for_error_and_print(file_to_check, errors)
    for i in found_errors.keys():
        errors_in_line = found_errors[i]
        print("Errors in line " + str(i))
        for one_error in errors_in_line:
            print(one_error)


def task_make(parsed_args, compilable_tex_files):
    """
    Compile one or more tex files under the specifications of the cmd args.

    Exists with exit code 1 if one or more tex files could not be translated.

    :param parsed_args: dict
        The parameter from the arg parser.
    :param compilable_tex_files:
        List of tex files that can be compiled.
    """
    is_verbose = parsed_args["verbose"]
    do_bib = parsed_args["bib"]
    do_produce_status_list = parsed_args["listStatus"]
    num_compile_attempts = parsed_args["number"]
    if parsed_args["all"]:
        if parsed_args["texfile"] != filename_stem(Configuration.get_config().main_tex):
            bash.print_warning("Tex file (" + str(parsed_args["texfile"]) + ")"
                               + " and --all specified. Tex file will be ignored.")
        all_compile_results = []
        for one_tex_file in compilable_tex_files:
            one_compile_result = make_pdf_with_substitute_defaults(one_tex_file,
                                                                   num_compile_attempts=num_compile_attempts,
                                                                   verbose=is_verbose,
                                                                   do_bib=do_bib)
            all_compile_results.append(one_compile_result)
            if do_produce_status_list:
                status = bash.Status.UNKNOWN
                status_message = None
                if one_compile_result.outcome == latex.Outcome.SUCCESS:
                    status = bash.Status.SUCCESS
                elif one_compile_result.outcome == latex.Outcome.ABORTED:
                    status = bash.Status.ERROR
                    status_message = "ABORTED"
                elif one_compile_result.outcome == latex.Outcome.FAILURE:
                    status = bash.Status.ERROR
                bash.print_status(one_tex_file, status, status_message)

        failed_compiles = [one_compile_result.filename for one_compile_result in all_compile_results
                           if one_compile_result.outcome != latex.Outcome.SUCCESS]
        num_fail_compile = len(failed_compiles)
        if num_fail_compile > 0:
            failed_compiles_list = ""
            for one_failed_compile in failed_compiles:
                failed_compiles_list += "\t- " + one_failed_compile + "\n"
            distance_to_previous_print = ""
            if is_verbose or do_produce_status_list:
                distance_to_previous_print += "\n"
            bash.print_error(distance_to_previous_print
                             + "Failed to compile all files. "
                             + str(num_fail_compile) + "/" + str(len(compilable_tex_files))
                             + " could not be compiled.\n" + failed_compiles_list[:-1])
            sys.exit(1)
    else:
        tex_file = parsed_args["texfile"] + ".tex"
        compile_one_tex(tex_file, do_bib=do_bib, is_verbose=is_verbose, num_compile_attempts=num_compile_attempts)


def task_free_compile_cmake(parsed_args):
    """
    Compile a tex files under the specifications of the cmd args.

    :param parsed_args:dict
        The parameter from the arg parser.
    """
    is_verbose = parsed_args["verbose"]
    do_bib = parsed_args["bib"]
    tex_file = parsed_args["maintex"]
    num_compile_attempts = parsed_args["number"]
    compile_one_tex(tex_file, do_bib=do_bib, is_verbose=is_verbose, num_compile_attempts=num_compile_attempts)


def select_task(parsed_args, compilable_tex_files):
    """
    Calls the respective function for the task as specified by the cmd args.

    :param parsed_args:dict
        The parameter from the arg parser.
    :param compilable_tex_files: list
         Individually compilable tex files in the current directory.
    """
    task = parsed_args["task"]
    if task == "make":
        task_make(parsed_args, compilable_tex_files)
    if task == "cmake":
        task_free_compile_cmake(parsed_args)
    elif task == "lang":
        task_lang_command(parsed_args)
    elif task == "conf":
        task_write_conf(parsed_args)
    elif task == "bibfile":
        task_make_bib_file(parsed_args)
    elif task == "clean":
        task_clean_up(parsed_args)
    elif task == "git":
        task_git(parsed_args)


def main():
    """
    Main entry of program. Set arguments for argument parser, loads the settings and
    searches current directory for compilable tex files. At the end prints help or calls
    the task selector function to perform one of the tasks.
    """
    parser = ap.ArgumentParser()
    root_logger = logging.getLogger()
    root_logger.handlers.append(logging.StreamHandler(sys.stdout))
    config = load_settings()
    if not config.log_to_console:
        logging.getLogger().removeHandler(logging.getLogger().handlers[1])
    compilable_tex_files = util.find_compilable_tex(".")
    if config.main_tex not in compilable_tex_files:
        compilable_tex_files.append(config.main_tex)
    compilable_tex_files_names = [os.path.splitext(f)[0] for f in compilable_tex_files]
    if len(compilable_tex_files_names) == 0:
        compilable_tex_files_names.append(NO_TEX_FOUND_WARNING)
    # Reusable arguments
    parent_with_tex_file = ap.ArgumentParser(add_help=False)
    parent_with_tex_file.add_argument("--maintex",
                                      type=lambda x: ArgumentParserExtensions.is_valid_file(x, parent_with_tex_file),
                                      default=config.main_tex,
                                      help="Specify main tex file")
    parent_with_bib = ap.ArgumentParser(add_help=False)
    parent_with_bib.add_argument("-b", "--bib", action="store_const", const=True, default=False,
                                 help="Compile bibliography.")
    parent_force_action = ap.ArgumentParser(add_help=False)
    parent_force_action.add_argument("-f",
                                     action="store_true",
                                     help="Force action "
                                          + "(do not ask for confirmation when overwriting or deleting files).")
    verbose_print = ap.ArgumentParser(add_help=False)
    verbose_print.add_argument('-v', '--verbose', action="store_const", const=True, default=False,
                               help="Activate verbose mode.")
    parent_number_ignore_latex_error = ap.ArgumentParser(add_help=False)
    parent_number_ignore_latex_error.add_argument('-n', '--number', type=int,
                                                  default=config.number_ignore_latex_compile_errors,
                                                  help="The number of attempts to ignore latex errors.")
    # Parsers for tasks
    task_parsers = parser.add_subparsers(dest='task')
    # Make config file
    # write_conf_parser =
    task_parsers.add_parser('conf',
                            parents=[parent_force_action],
                            help=f"Create configuration file ({CONF_FILE}).")
    # Language check
    lang_check_parser = task_parsers.add_parser('lang', help='Search for language errors.')
    lang_check_parser.add_argument("file",
                                   type=lambda x: ArgumentParserExtensions.is_valid_file(x, lang_check_parser),
                                   help="File to check for language errors.")
    error_identifiers = config.language_errors.identifiers
    if len(error_identifiers) > 0 and len(error_identifiers) != len(set(error_identifiers)):
        logging.warning("Multiple text errors with the same identifier.")
    for one_error_identifier in set(error_identifiers):
        associated_error_message = config.language_errors.get_error_for_identifier(one_error_identifier).error_message
        help_message = f"Exclude error: {associated_error_message}."
        lang_check_parser.add_argument(f"--{one_error_identifier.strip()}",
                                       action="store_const", const=True, default=False,
                                       help=help_message)
    # Free make pdf
    # make_parser =
    task_parsers.add_parser('cmake',
                            parents=[parent_with_tex_file, parent_with_bib, verbose_print,
                                     parent_number_ignore_latex_error],
                            help='Translate document.')
    # Make pdf
    make_compile_parser = task_parsers.add_parser('make',
                                                  parents=[parent_with_bib, verbose_print,
                                                           parent_number_ignore_latex_error],
                                                  help='Translate document.')
    make_compile_parser.add_argument("texfile",
                                     nargs="?",
                                     default=filename_stem(config.main_tex),
                                     choices=compilable_tex_files_names,
                                     help="File to compile.")
    make_compile_parser.add_argument("--all", action="store_const", const=True, default=False,
                                     help="Translate all independent tex files in the directory.")
    make_compile_parser.add_argument("-l", "--listStatus", action="store_const", const=True, default=False,
                                     help="Produce a list of all compile outcomes.")
    # Clean
    clean_parser = task_parsers.add_parser('clean',
                                           parents=[parent_force_action],
                                           help='Cleans the directory.')
    clean_parser.add_argument("-d", type=int, default=config.clean_depth,
                              help="Set the maximal depth into sub-folders for looking for files to delete.")
    clean_parser.add_argument("-l", action="store_const", const=True, default=False,
                              help="Only list files and do not delete.")
    # Make bibliography file
    make_bib_file = task_parsers.add_parser('bibfile',
                                            parents=[parent_force_action],
                                            help='Create bib file from citations in aux file.')
    make_bib_file.add_argument("-a", "--auxFile",
                               type=lambda x: ArgumentParserExtensions.is_valid_file(x, make_bib_file),
                               default=config.main_aux,
                               help="The aux file.")
    make_bib_file.add_argument("-b", "--bibsFolder",
                               type=lambda x: ArgumentParserExtensions.is_valid_directory(x, make_bib_file),
                               default=config.bibs_folder,
                               help="The path to the folder containing the bib files to search for entries.")
    make_bib_file.add_argument("-o", "--output", type=str,
                               default=config.main_bibliography_file,
                               help="The file to write the bibliography entries into.")
    # Git help
    git_parser = task_parsers.add_parser("git",
                                         help="Help with git related operations.")
    git_action_parsers = git_parser.add_subparsers(dest='action')
    git_action_parsers.add_parser("ignore",
                                  help="Add default clean extensions to gitignore.")
    credentials_parser = git_action_parsers.add_parser("credentialUsername",
                                                       help="Add credential username to local git.")
    credentials_parser.add_argument("username",
                                    type=str,
                                    help="The username to use.")
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    args = vars(args)
    if args["task"] is not None:
        select_task(args, compilable_tex_files)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



"""
Module for shell interaction with git.
"""

import subprocess


class GitInteraction:
    """
    Class to allow git shell commands to be run.
    """

    CONFIG = 1

    @staticmethod
    def execute_bash_git_cmd(cmd_type, parameter):
        """
        Execute a specified git sub-command with the given parameters.

        :param cmd_type: int
            One of the command types specified by this class.
        :param parameter: str
            The parameter(s) for the git sub-command.
        :return: str
            The errors if any occurred or None.
        """
        if cmd_type == GitInteraction.CONFIG:
            sub_command = "config"
        else:
            raise ValueError("The command type has to be one defined by the GitInteraction class.")
        git_cmd_process = subprocess.Popen(f"git {sub_command} {parameter}", shell=True, stdout=subprocess.PIPE)
        (output, error) = git_cmd_process.communicate()
        git_cmd_process.wait()
        return error

    @staticmethod
    def get_git_config_credential_set_username_parameter_string(username):
        """
        Concatenate the username with the git config sub-command parameters.
        :param username: str
            The username to use.
        :return: str
            The string '--local credential.${{remote}}.username <username>' where <username> is replaced with
            the passed parameter.
        """
        return f"--local credential.${{remote}}.username {username}"



"""
Compiling tex files and processing bibliography (bib) files.
"""

import subprocess
import sys

import latex_suite.util


class Outcome:
    """
    The outcome of compiling or processing latex and related files.
    """

    UNKNOWN = -1
    SUCCESS = 0
    FAILURE = 1
    ABORTED = 2
    FILE_NOT_FOUND = 3


class ProcessRunnerWithOutput:
    """
    Class to run a cmd line program on a positional parameter (e.g. a file) and collect the output.
    """

    def __init__(self, command, parameter, verbose=False):
        self._command = command
        self._parameter = str(parameter)
        self._verbose = verbose
        self._process = subprocess.Popen(self._command + " " + self._parameter, shell=True,
                                         stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        self._total_output = ""

    @property
    def output(self):
        """
        :return:
        The output written to stdout by the pdf compiler engine.
        """
        return self._total_output

    @property
    def command(self):
        """
        The command/process.
        :return: str
            The command/process.
        """
        return self._command

    @property
    def parameter(self):
        """
        The parameter of the command/process.
        :return:
            The parameter of the command/process.
        """
        return self._parameter

    def write_to_stdin(self, input_bytes):
        """
        Write bytes to sdtin.
        :param input_bytes: bytes
            The input to write.
        """
        self._process.stdin.write(input_bytes)
        self._process.stdin.flush()

    def write_newline(self):
        """
        Write newline to stdin.
        """
        self.write_to_stdin(b'\n')

    def __iter__(self):
        """
        :return:
        An iterator over the bytes of the stdout from the compiler.
        """
        self._stdout_iterator = iter(lambda: self._process.stdout.read(1), b'')
        return self

    def __next__(self):
        """
        Get the next char from stdout until the process the finished or stdin gets closed.
        The char is also added to the entire output (see :func:`~LatexBashCompile.output`).

        :return: str
            The next char.
        """
        if self._process.stdin.closed:
            raise StopIteration
        byte_char = next(self._stdout_iterator)
        char = byte_char.decode("utf-8")
        self._total_output += char
        if self._verbose:
            sys.stdout.write(char)
        return char


class BibBashProcessing(ProcessRunnerWithOutput):
    """
    Process citations of a latex document.
    """

    def __init__(self, engine, tex_file, verbose=False):
        super().__init__(engine, BibBashProcessing.get_bib_parameter(engine, tex_file), verbose)
        self._verbose = verbose

    @staticmethod
    def get_bib_parameter(engine, tex_file):
        """
        Return the parameter of the bib engine to process citations.

        :param engine: str
            The engine.
        :param tex_file: str
            The name of the text file.
        :return: str
            The parameter for the bib engine.
        """
        bib_filename = None
        if engine == "biber":
            bib_filename = latex_suite.util.filename_stem(tex_file)
        elif engine == "bibtex":
            bib_filename = latex_suite.util.filename_stem(tex_file) + ".aux"
        return bib_filename

    def run_processing(self):
        """
        Process the citations.
        :return: int
            The outcome indicator.
        """
        compile_outcome = Outcome.SUCCESS
        for _ in iter(self):
            pass
        if self._verbose:
            sys.stdout.write("\n")
        sys.stdout.flush()
        if "I couldn't open file name" in self._total_output:
            compile_outcome = Outcome.FILE_NOT_FOUND
        if "I found no \\bibstyle command" in self._total_output:
            compile_outcome = Outcome.FAILURE
        return compile_outcome


class LatexBashCompile(ProcessRunnerWithOutput):
    """
    Class to use bash tex compile tools to translate pdfs.
    """

    def __init__(self, engine, file, max_end_attempts, verbose=False):
        """
        Constructor

        :param engine:
            The pdf compiler engine.
        :param file:
            The tex file to translate.
        :param max_end_attempts:
            The maximal number of attempts to progress past a compile error.
        """

        super().__init__(engine, file, verbose)
        self._max_end_attempts = max_end_attempts
        self._verbose = verbose
        self._warning_progressions = 0
        self._error_progressions = 0

    @property
    def num_warnings(self):
        """
        :return: int
        The number of question mark warnings during the pdf compile process.
        """
        return self._warning_progressions

    @property
    def num_errors(self):
        """
        :return: int
        The number of asterisk errors during the pdf compile process.
        """
        return self._error_progressions

    def write_end_enter(self):
        """
        Passes \\end and newline to stdin for passing past a latex error that requires entering an
        \\end command.
        """
        self.write_to_stdin(b'\\end\n')

    def run_compile(self):
        """
        Compile the tex file with the specified engine.

        :return: LatexBashCompileResult
        The compile result with the outcome, output and number of errors and warnings.
        """
        compile_outcome = Outcome.SUCCESS
        end_attempts = 0
        start_of_new_line = True
        potential_file_not_found_idx = 0
        for char in iter(self):
            if start_of_new_line:
                if char == "?":
                    self.write_newline()
                    self._warning_progressions += 1
                elif char == "!":
                    potential_file_not_found_idx = 1
                elif char == "*":
                    if end_attempts >= self._max_end_attempts:
                        self._process.stdin.close()
                        compile_outcome = Outcome.ABORTED
                    else:
                        self.write_end_enter()
                        end_attempts += 1
                        self._error_progressions += 1
            elif potential_file_not_found_idx > 0:
                fnf_message = "! I can't find file"
                if char == fnf_message[potential_file_not_found_idx]:
                    potential_file_not_found_idx += 1
                else:
                    potential_file_not_found_idx = 0
                if potential_file_not_found_idx == len(fnf_message):
                    self._process.stdin.close()
                    compile_outcome = Outcome.FILE_NOT_FOUND
            start_of_new_line = False
            if char == "\n":
                start_of_new_line = True
        if self._verbose:
            sys.stdout.write("\n")
        sys.stdout.flush()
        if (compile_outcome != Outcome.ABORTED
                and "Fatal error occurred, no output PDF file produced!" in self._total_output):
            compile_outcome = Outcome.FAILURE
        return compile_outcome


class LatexBashCompileResult:
    """
    The results of a tex compilation.
    """

    def __init__(self, filename, outcome, output, num_warnings, num_errors):
        self.filename = filename
        self.outcome = outcome
        self.output = output
        self.num_warnings = num_warnings
        self.num_errors = num_errors
        self._bib_processing_result = None

    @property
    def bib_processing_result(self):
        """
        :return:
        BibBashProcessingResult
            The result of a bib processing.
        """
        return self._bib_processing_result

    @bib_processing_result.setter
    def bib_processing_result(self, result):
        """
        Set the result of a bib processing.
        :param result: BibBashProcessingResult
            The result.
        """
        self._bib_processing_result = result


class BibBashProcessingResult:
    """
    The results of a bib processing.
    """

    def __init__(self, filename, outcome, output):
        self.filename = filename
        self.outcome = outcome
        self.output = output


def make_pdf(file_to_translate, engine, bib_engine, max_end_attempts=1,
             verbose=False, do_bib=False, do_single_run=False):
    """
    Compile a tex file to a pdf.

    Performs one or several steps of the translation of a tex file to a pdf file.
    This can include single and several compilation runs as well as citation processing.
    The minimal run is one translation of the tex file (do_bib=False, do_single_run=True) and the maximal
    run is: tex translation, bib processing, tex translation, tex translation (do_bib=True, do_single_run=False).

    :param file_to_translate:
        The path of the tex file.
    :param engine: str
        The engine to use to translate the tex file.
    :param bib_engine: str
        The engine to use to process the citations.
    :param max_end_attempts: int
        The maximal number of attempts to try to pass past a latex translation error.
    :param verbose: bool
        If True print processing.
    :param do_bib: bool
        If True processes the citations with the citation engine.
    :param do_single_run:
        If True and do_bib is False one translation of the tex file.
        If True and do_bib is True: tex translation, citation processing, tex translation.
    :return: LatexBashCompileResult
        The compile result of the last compile with the outcome, output and number of errors and warnings.
        If the bibliography was processed the bib processing information is included.
    """
    compile_result = None
    bib_process_result = None
    if do_bib and not do_single_run or not do_bib and do_single_run or not do_bib and not do_single_run:
        compile_result = compile_file(engine, file_to_translate, max_end_attempts=max_end_attempts, verbose=verbose)
    if do_bib:
        bib_process_result = compile_bib(bib_engine, file_to_translate, verbose=verbose)
    if do_bib and not do_single_run:
        compile_result = compile_file(engine, file_to_translate, max_end_attempts=max_end_attempts, verbose=verbose)
    if not do_single_run:
        compile_result = compile_file(engine, file_to_translate, max_end_attempts=max_end_attempts, verbose=verbose)
    compile_result.bib_processing_result = bib_process_result
    return compile_result


def compile_bib(engine, bib_parameter, verbose=False):
    """
    Process citations.

    :param engine: str
        The engine to use to process the citations.
    :param bib_parameter: str
        The parameter for the bib engine to process the citations. This could be a file name or a file name without
        the file extension.
    :param verbose: bool
         If True print processing.
    :return: BibBashProcessingResult
        The result of the citation processing.
    """
    processor = BibBashProcessing(engine, bib_parameter, verbose)
    outcome = processor.run_processing()
    return BibBashProcessingResult(BibBashProcessing.get_bib_parameter(processor.command, processor.parameter),
                                   outcome, processor.output)


def compile_file(engine, file, max_end_attempts=1, verbose=False):
    """
    Compiles a tex file with the specified engine.

    :param engine:
        The pdf compiler engine.
    :param file:
        The tex file to translate.
    :param max_end_attempts:
        The maximal number of attempts to progress past a compile error.
    :param verbose: bool
        If true prints engines output.
    :return: LatexBashCompileResult
        The compile result with the outcome, output and number of errors and warnings.
    """
    compiler = LatexBashCompile(engine, file, max_end_attempts, verbose)
    outcome = compiler.run_compile()
    return LatexBashCompileResult(file, outcome, compiler.output, compiler.num_warnings, compiler.num_errors)


def get_warnings(engine, file):
    """
    Get all warnings of a tex translation.

    :param engine:
        The pdf compiler engine.
    :param file:
        The tex file to translate.
    :return: list
        All warnings that occurred during the compilation.
    """
    compile_file(engine, file)
    compile_result = compile_file(engine, file)
    warnings = []
    warning_indicators = ["multiply defined", "overfull", "undefined", "Warning: Reference", "Warning:"]
    for one_line in compile_result.output.split("\n"):
        if any(indicator in one_line for indicator in warning_indicators):
            warnings.append(one_line)
    return warnings



"""
Creation of bibliography (bib) file with required entries subset of all bib entries in several files.
"""

import os

from pybtex import database as bib_parser


def get_all_references_from_aux(aux_file_path):
    """
    Determine citation keys from aux file.

    :param aux_file_path:
        The path of the aux file.
    :return:
    [str]
        The list of all citation keys.
    """
    all_references = set()
    with open(aux_file_path, "r") as aux_file:
        for line in aux_file:
            if line.startswith("\\abx@aux@cite") or line.startswith("\\citation"):
                reference = line.strip().replace("\\abx@aux@cite{", "").replace("\\citation{", "").replace("}", "")
                for one_reference in reference.split(","):
                    all_references.add(one_reference)
    return all_references


def add_entries_from_file(citation_keys, bib_folder_path, bib_db, fields_to_remove, errors_list):
    """
    Adds all references from the files in the bib folder which have a citation key in the specified set
    to the db of references.

    Only files with the extension 'bib' are considered.

    :param citation_keys: [str]
        List of citation keys to look for.
    :param bib_folder_path:
        The folder with bib files path.
    :param bib_db: bib_parser.BibliographyData
        A database to add references to.
    :param fields_to_remove: [str]
        Reference entries' fields that should be removed before adding an entry to the db.
    :param errors_list:
        A list to add any occurring errors to
    """
    for one_file in os.listdir(bib_folder_path):
        if one_file.endswith(".bib"):
            bib_file_bib_db = bib_parser.parse_file(os.path.join(bib_folder_path, one_file))
            for key, entry in bib_file_bib_db.entries.items():
                try:
                    if key in citation_keys:
                        for one_field_to_remove in fields_to_remove:
                            if one_field_to_remove in entry.fields:
                                entry.fields.pop(one_field_to_remove)
                        for field_key in list(entry.fields.keys()):
                            if "\\&" in entry.fields[field_key]:
                                entry.fields[field_key] = entry.fields[field_key].replace("\\&", "&")
                        bib_db.add_entry(key, entry)
                except bib_parser.BibliographyDataError as ex_bib:
                    errors_list.append(ex_bib)


def create_bib_file(aux_file_path, bib_folder_path, out_file_path, fields_to_remove=None):
    """
    Determines all citation keys from an aux file and creates a bib file with all
    the references for these citation keys.

    :param aux_file_path:
        The aux file path.
    :param bib_folder_path:
        The folder with bib files path.
    :param out_file_path:
        The path of the bib file to write to.
    :param fields_to_remove: [str]
        Reference entries' fields that should be removed before adding an entry to the bib file.
    :return:
    [bib_parser.BibliographyDataError]
        A list of errors.
    """
    if fields_to_remove is None:
        fields_to_remove = []
    errors = []
    citation_keys = get_all_references_from_aux(aux_file_path)
    full_bib_db = bib_parser.BibliographyData()
    add_entries_from_file(citation_keys, bib_folder_path, full_bib_db, fields_to_remove, errors)
    full_bib_db.to_file(out_file_path)
    return errors



"""
Some utility functions for the configuration of the suite and file operations.
"""

import os
import re

import yaml

from latex_suite.search_language_errors import TextErrorType, ErrorList


FILE_EXTENSION_PATTERN = re.compile(r"\*\.[a-zA-Z0-9]+")
DOCUMENT_CLASS_NOT_IN_COMMENT_PATTER = re.compile(r"[ a-zA-Z0-9_]*%[ a-zA-Z0-9_]*\\documentclass[ a-zA-Z0-9_]*")

not_working_errors = [
    TextErrorType("No text between sections.",
                  r"(section|chapter|paragraph){[\w\" \",:]+}[\n\" \"\t]+"
                  + r"(\\label{[\w\" \",:]*})*[\n\" \"\t]+\\(sub)*(section|paragraph)")
    ]

verbose_errors = [
    TextErrorType("Capital in sentence.", r"[a-zA-z0-9_\(\)\[\]\{\}:]\ [A-Z]")
    ]


class Configuration:
    """
    Configuration for standard options.
    """

    engine: str
    bib_engine: str
    main_tex: str
    clean_extensions: list[str]
    clean_depth: int
    bibs_folder: str
    main_bibliography_file: str
    bibliography_fields_to_remove: list[str]
    number_ignore_latex_compile_errors: int
    language_errors: ErrorList
    log_to_console: bool

    DEFAULT_CONF = {
        "engine": ("engine", "pdflatex"),
        "bib_engine": ("bib_engine", "biber"),
        "main_tex": ("main_tex", "main.tex"),
        "clean_extensions": ("clean_file_extensions", [".log", ".aux", ".bbl", ".blg"]),
        "clean_depth": ("clean_depth", 0),
        "bibs_folder": ("bib_files_folder", "."),
        "main_bibliography_file": ("main_bibliography_file", "bibliography.bib"),
        "bibliography_fields_to_remove": ("bibliography_fields_to_remove", ["abstract", "file", "keywords", "url"]),
        "number_ignore_latex_compile_errors": ("number_ignore_latex_compile_errors", 1),
        "log_to_console": ("log_to_console", False),
        "language_errors": ("language_errors",
                            ErrorList([
                                TextErrorType("'An' not in front of vowel.", r"[\.\ ][aA]n\ [^aeiouAEIOU]"),
                                TextErrorType("'A' in front of vowel.", r"[\.\ ][aA]\ [aeiouAEIOU]"),
                                TextErrorType("Extra space.", r"[^\ \n%][\ ]{2,}[^\ \n]"),
                                TextErrorType("Named reference with lower case",
                                              r"\b[a-z]\w*\b[\ ~]\\ref\{", ["and", "to"], " |~"),
                                TextErrorType("Doubled word.", r"\b(\w+)\s+\1\b"),
                                TextErrorType("Citation with name and not shortcite",
                                              r"(\b)( et al.){1}~\\cite\{ref:\1", identifier="sc"),
                                TextErrorType("Every sentence in a line",
                                              r"[^iegst0-9\"al\"]\.[^\n]*\.", identifier="spl")]))
    }
    IDX_TAG = 0
    IDX_VALUE = 1

    _instance = None

    def __init__(self, loaded_yaml=None):
        """
        Constructor that uses the yaml config file contents to set the configuration and sets anything that is
        not specified to a default value.
        :param loaded_yaml:
        """
        if loaded_yaml is None:
            loaded_yaml = {}
        self._set_default()
        for one_key in self.DEFAULT_CONF.keys():
            one_tag = self.DEFAULT_CONF[one_key][self.IDX_TAG]
            if one_tag in loaded_yaml:
                internal_key = "_" + one_key
                self.__setattr__(internal_key, loaded_yaml[one_tag])

    def __new__(cls, *args, **kwargs):
        """
        Ensures that there is only one resource class.
        """

        if cls._instance is None:
            cls._instance = super(Configuration, cls).__new__(cls)
        return cls._instance

    def _set_default(self):
        """
        Set all known settings to default.
        """
        for one_key in self.DEFAULT_CONF.keys():
            internal_key = "_" + one_key
            self.__setattr__(internal_key, self.DEFAULT_CONF[one_key][self.IDX_VALUE])

    def __getattribute__(self, item):
        """
        If the key is a default setting retrieve that entry otherwise call super.
        :param item:
        :return:
        """
        if item != "DEFAULT_CONF" and item in self.DEFAULT_CONF.keys():
            value = self._get_conf_value(item)
        else:
            value = super().__getattribute__(item)
        return value

    def __setattr__(self, key, value):
        """
        If the key is a default setting raise AttributeError since these cannot be changed; otherwise call super.
        :param key:
        :param value:
        :return:
        """
        if key != "DEFAULT_CONF" and key in self.DEFAULT_CONF.keys():
            raise AttributeError("can't set attribute")
        else:
            super().__setattr__(key, value)

    @classmethod
    def get_config(cls):
        """
        :return: Configuration
        The configuration if it has been initialised before and None otherwise.
        """
        return cls._instance

    def _get_conf_value(self, key):
        """
        Property step in for configuration settings.
        :param key:
        :return:
        """
        value = self.__getattribute__("_" + key)
        return value

    @property
    def main_aux(self):
        """
        The aux file associated with the main tex file.
        :return:
        str
            '<main tex file name without extension>.aux'
        """
        return filename_stem(self.main_tex) + ".aux"

    @staticmethod
    def write_default(file):
        """
        Write the default configuration to a yaml file.
        :param file:
        The path of the file to write to.
        """
        default_config = {}
        for one_key in Configuration.DEFAULT_CONF:
            tag = Configuration.DEFAULT_CONF[one_key][Configuration.IDX_TAG]
            value = Configuration.DEFAULT_CONF[one_key][Configuration.IDX_VALUE]
            default_config[tag] = value
        yaml.Dumper.add_multi_representer(ErrorList, ErrorList.to_yaml)
        with open(file, "w") as config_file:
            yaml.dump(default_config, config_file)


def remove_files(file_extensions, directory=None, depth=0, only_list=False, force_remove=False):
    """
    Removes all files with matching file extensions.

    :param file_extensions: [str]
        The list of file extensions (including dot (.)), e.g. [".log"]
    :param directory:
        The root directory to start the delete process.
    :param depth: int
        The recursion depth to go to for deleting files. 0 means only the directory, 1 includes the subdirectories,
        2 includes the subdirectories of the subdirectories and so on.
    :param only_list:
        Does not delete the files. Only list them.
    :param force_remove:
        Just removes the files without prompting for input.
    """

    num_files = 0
    num_files_removed = 0
    if directory is None:
        directory = "."
    for dir_entry in os.listdir(directory):
        dir_entry_path = os.path.join(directory, dir_entry)
        if os.path.isfile(dir_entry_path):
            if filename_extension(dir_entry) in file_extensions:
                num_files += 1
                if only_list:
                    print(dir_entry_path)
                else:
                    response = None
                    if force_remove:
                        response = "y"
                    while response is None or (response != "y" and response != "n"):
                        text = f"Please enter 'y' or 'n' to remove or not remove '{dir_entry_path}', respectively:"
                        if response is None:
                            text = f"Do you want to remove '{dir_entry_path}? y/n:"
                        response = input(text)
                    if response == "y":
                        print("Removing '" + str(str(dir_entry_path)) + "'.")
                        os.remove(dir_entry_path)
                        num_files_removed += 1
        elif os.path.isdir(dir_entry_path) and depth > 0:
            num_files_sub_dirs, num_files_removed_sub_dirs = remove_files(file_extensions, dir_entry_path,
                                                                          depth=depth-1, only_list=only_list,
                                                                          force_remove=force_remove)
            num_files += num_files_sub_dirs
            num_files_removed += num_files_removed_sub_dirs
    return num_files, num_files_removed


def find_compilable_tex(directory_path):
    """
    Find tex files that can be compiled on their own.

    A text file is considered compilable on its own if the file extension is tex and if the first
    non comment line contains a \\documentclass command.
    :param directory_path:
        The directory to look for compilable tex files.
    :return:
    [str]
        List of relative paths of files which are compilable.
    """
    tex_files = [f for f in os.listdir(directory_path)
                 if os.path.isfile(os.path.join(directory_path, f))
                 and os.path.splitext(f)[1] == ".tex"]
    compilable_tex_files = []
    for one_tex_file in tex_files:
        with open(one_tex_file, "r") as tex_file:
            continue_searching = True
            while continue_searching:
                try:
                    line = next(tex_file)
                except StopIteration:
                    continue_searching = False
                    line = "EOF"
                if "\\documentclass" in line and not DOCUMENT_CLASS_NOT_IN_COMMENT_PATTER.search(line):
                    compilable_tex_files.append(one_tex_file)
                if (not ("\\documentclass" in line and not DOCUMENT_CLASS_NOT_IN_COMMENT_PATTER.search(line)
                         or "".join(line.split()).startswith("%")
                         or len("".join(line.split())) == 0)):
                    continue_searching = False
    return compilable_tex_files


def add_files_and_extensions_to_file(all_patters_to_add, path):
    """
    Adds all file patterns to the file in path.

    File extension patterns are appended to the longest list of file extension
    patterns or the end of the file. All other patterns are added to the end of the
    file. The list of patterns does not need newlines these are added in. The patterns
    can be files or file extension, e.g. ['*.jpg', 'some.pdf'].

    :param all_patters_to_add: [str]
        The list of patterns.
    :param path: str
        The path to the file.
    """

    with open(path, "r") as file:
        lines = file.readlines()
    all_contents = "".join(lines)
    extension_insert_location = find_file_extension_insert_location(lines)
    file_extension_matcher = re.compile(FILE_EXTENSION_PATTERN)
    for one_pattern_to_add in all_patters_to_add:
        if one_pattern_to_add not in all_contents:
            if file_extension_matcher.match(one_pattern_to_add):
                lines.insert(extension_insert_location, one_pattern_to_add + "\n")
                extension_insert_location += 1
            else:
                lines.insert(len(lines), one_pattern_to_add + "\n")
                if extension_insert_location == len(lines):
                    extension_insert_location += 1
    with open(path, "w") as f:
        contents = "".join(lines)
        f.write(contents)


def find_file_extension_insert_location(lines):
    """
    Find the longest sequence of lines containing file extension entries and choose the end
    of this as the insertion location. If there are no file extension lines returns end of file
    as index.
    :param lines: [str]
        The file as a list of lines.
    :return: int
        The index after the longest sequence or the end of file (len of lines list).
    """

    insert_location = len(lines)
    current_sequence_length = -1
    longest_sequence_length = 0
    file_extension_matcher = re.compile(FILE_EXTENSION_PATTERN)
    for i in range(len(lines)):
        one_line = lines[i]
        if file_extension_matcher.match(one_line):
            if current_sequence_length > 0:
                current_sequence_length += 1
            else:
                current_sequence_length = 1
        elif current_sequence_length > 0:
            if current_sequence_length > longest_sequence_length:
                insert_location = i
                longest_sequence_length = current_sequence_length
            current_sequence_length = -1
    return insert_location


def filename_stem(filename):
    """
    Returns the name of the file without the file extension.
    :param filename: The full name of the file with file extension.
    :return: str
        The name of the file.
    """

    return os.path.splitext(filename)[0]


def filename_extension(filename):
    """
    Returns the file extension of the file.
    :param filename: The full name of the file with file extension.
    :return: str
        The file extension.
    """

    return os.path.splitext(filename)[1]



from distutils.core import setup
from os import path
import pathlib


root_dir = pathlib.Path(__file__).parent
with open(path.join(root_dir, 'requirements.txt')) as requirements_file:
    all_lines = requirements_file.read().split('\n')
    all_requirements = [r.strip() for r in all_lines]

setup(
    name="latexsuite",
    version="0.9.1",
    description="Support your bash needs when working with latex projects in git repositories.",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/jbuerman/latexsuite",
    author="Jan Buermann",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
    packages=["latex_suite"],
    include_package_data=True,
    install_requires=all_requirements,
    entry_points={"console_scripts": ["latexsuite=latex_suite.latex_suite:main"]},
)
